/* MY LINUX CODE TO DEFINE NEW SYSCALL: getCPUinfo THAT CHECKS THE CPU INFO,
 *from asm/processor.h, Control Reg. CR4 & Intel CPUID instr-n
 * I add some more code to make it print to arbitrary tty
 *
 * lowpowered by: canturk isci / canturk@princeton.edu / HOBAA
 */

/* I AM USING SYSCALL # 222 and UP FOR MY SYSCALLs AS  ASM/UNISTD.H GOES
  *UP TO 221 */

/*
COMPILE WITH:
gcc -DMODULE -D__KERNEL__ -DLINUX -Wall -I/lib/modules/`uname -r`/build/include
-c LKM_CPUinfo.c -o LKM_CPUinfo.o
(You can get rid of some of the compile -D directives by defining them in code, but who cares?)
*/

/* Standard in kernel modules */
#include <linux/kernel.h>
#include <linux/module.h>

/* For syscalls*/
#include <sys/syscall.h>
#include <linux/sys.h>

/* For tty */
#include <linux/sched.h> /* For current */
#include <linux/tty.h> /* For the tty declarations */

/* Deal with CONFIG_MODVERSIONS */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <stdio.h>
/*#include <string.h> --> DOESN't work with LKMs!! u need the below one:*/
#include <linux/string.h> /* for memcpy */

/* FOR CPUID & CR4 STUFF */
#include <asm/processor.h>
/*extern void check_bugs(void);
#include <asm/bugs.h>
#include <asm/pgtable.h>*/

/* DEFINE THE NUMBERs I CHOSE FOR MY SYSCALLs */
/* (May be made user enterable on command line while insmod */
#define MY_SYSCALL_NO1 222

/* INCLUDE SYSTEM CALL TABLE to change it*/
extern void *sys_call_table[];

/* MY syscall */
asmlinkage static int sys_getCPUinfo();

/* Old Syscall table references for the new syscall numbers */
void *old_reference222;

/* MY ASSEMBLY CODE TO READ CR4: */
#define rdCR4(CR4value) \
        __asm__ __volatile__("movl %%cr4, %0\n" /* mov size=long cr4 data into the reg for cr4val*/\
        :"=r"(CR4value)) /* o/p = CR4value */

/* MY MACRO TO TEST A BIT IN A 32 bit DATA */
/*return 0 if bit at bitnum = 0; 1 if bit = 1 */
#define testbit32(testdata,bitnum)\
        ((((0x00000000 | (1 << bitnum)) & testdata)==0) ? 0 : 1)

/* FLAG BIT DEFINITIONS FOR EDX (BASIC CPUID LEVEL 1) */
/*(mostly copied from Intel's CPUID3.c ) */
#define FPU_FLAG 0x0001
#define VME_FLAG 0x0002
#define DE_FLAG 0x0004
#define PSE_FLAG 0x0008
#define TSC_FLAG 0x0010
#define MSR_FLAG 0x0020 /*bit 5 */
#define PAE_FLAG 0x0040
#define MCE_FLAG 0x0080
#define CX8_FLAG 0x0100
#define APIC_FLAG 0x0200
/* bit 10 reserved */
#define SEP_FLAG 0x0800
#define MTRR_FLAG 0x1000
#define PGE_FLAG 0x2000
#define MCA_FLAG 0x4000
#define CMOV_FLAG 0x8000 /* bit 15 */
#define PAT_FLAG 0x10000
#define PSE36_FLAG 0x20000
#define PSNUM_FLAG 0x40000
#define CLFLUSH_FLAG 0x80000
/* bit 20 reserved */
#define DTS_FLAG 0x200000
#define ACPI_FLAG 0x400000
#define MMX_FLAG 0x800000
#define FXSR_FLAG 0x1000000
#define SSE_FLAG 0x2000000  /* bit 25 */
#define SSE2_FLAG 0x4000000
#define SS_FLAG 0x8000000
#define HTT_FLAG 0x10000000
#define TM_FLAG 0x20000000
/* bit 30 reserved */
#define SBF_FLAG 0x80000000


/* I SORT OF COPY FROM /asm/processor.h MY CPUID IMPLEMENTATION: */
inline void MYcpuid(int op, int *eax, int *ebx, int *ecx, int *edx)
{
  __asm__("cpuid"
 : "=a" (*eax), /* outputps */
		  "=b" (*ebx),
		  "=c" (*ecx),
		  "=d" (*edx)
		: "0" (op));  /* input */
}


/* Function to print to the tty the current task uses */
/*Apparently, only prints up to 4096 bytes, ignores the rest*
 *tty->driver.flush_buffer can be used to flush it, but it  *
 *doesn't make it write the flushed portion to the tty device*/
void print_toTTY(char *str)
  {
    struct tty_struct *my_tty;

    /* The tty for the current task */
    my_tty = current->tty;

    /* If my_tty is NULL, it means that the current task
    * has no tty you can print to */

    if (my_tty != NULL) {

      /* my_tty->driver is a struct which holds the tty's
       * functions, one of which (write) is used to
       * write strings to the tty. It can be used to take
       * a string either from the user's memory segment
       * or the kernel's memory segment.
       *
       * The function's first parameter is the tty to
       * write to, because the  same function would
       * normally be used for all tty's of a certain type.
       * The second parameter controls whether the
       * function receives a string from kernel memory
       * (false, 0) or from user memory (true, non zero).
       * The third parameter is a pointer to a string,
       * and the fourth parameter is the length of
       * the string.
       */
      (*(my_tty->driver).write)(
         my_tty, /* The tty itself */
         0, /* We don't take the string from user space */
         str, /* String */
         strlen(str));  /* Length */
      (*(my_tty->driver).write)(
         my_tty,
         0,
         "\015\012",
         2);
    }
  }


/* DEFINITION OF MY SYSCALL */
asmlinkage static int sys_getCPUinfo()
{

  /* VAR-S FOR asm/processor.h data */
  extern struct cpuinfo_x86 boot_cpu_data;
  /* For some reason, I have to define it again in my code
   *although it's already defined in processor.h */
  int  cpuid_level;
  char vendor_id[16];
  char model_id[64];
  int  cache_size;

  /* VAR-S FOR CPUID */
  unsigned int eax_in, eax, ebx, ecx, edx; /* var-s to pass to GPRs*/
  char vendor_id2[13];/*Place to hold 12 char vendor id + 1 for null */
  unsigned int * pointer_to_char;/* pointer to pass to cpuid */
  char processor_type[4][32] = {
         "Original OEM Processor",
         "OverDrive Processor",
         "Dual Processor",
         "I Dunno (Intel Reserved"};
  unsigned int extended_family_no, family_code,
               extended_model_no, model_no,
               stepping_id;
  char brand_table[][64] =
  { /*0x00*/ " Unsupported",
    /*0x01*/ " Genuine Intel(R) Celeron(R) processor",
    /*0x02*/ " Genuine Intel(R) Pentium(R) III processor",
    /*0x03*/ " Genuine Intel(R) Pentium(R) III Xeon(TM) processor",
    /*0x04*/ " Genuine Intel(R) Pentium(R) III processor",
    /*0x05*/ " BAD BRAND ID",
    /*0x06*/ " Genuine Mobile Intel(R) Pentium(R) III Processor-M",
    /*0x07*/ " Genuine Mobile Intel(R) Celeron(R) processor",
    /*0x08*/ " Genuine Intel(R) Pentium(R) 4 processor",
    /*0x09*/ " Genuine Intel(R) Pentium(R) 4 processor",
    /*0x0A*/ " Genuine Intel(R) Celeron(R) processor",
    /*0x0B*/ " Genuine Intel(R) Xeon(TM) processor",
    /*0x0C*/ " Genuine Intel(R) Xeon(TM) processor MP",
    /*0x0D*/ " BAD BRAND ID",
    /*0x0E*/ " Genuine Mobile Intel(R) Pentium(R) 4 processor",
    /*0x0F*/ " Genuine Mobile Intel(R) Celeron(R) processor" };

  /* VAR-S FOR CONTROL REGISTERS' DATA */
  unsigned long CR4contents; /* CONTENTS OF CR4 REG */
  /*My Beautiful 3D structure to print info about CR4 bits */
  /*Each row is for one bit CR4, starting from bit0 */
  /*Prints left message if bit not set, right one if bit set */
  char CR4_bitmessages[][2][128] = {
         {"VME not Set", "VME Set"},       /* 0 */
         {"PVI not Set", "PVI Set"},       /* 1 */
         {"TSD not Set (RDTSC Can be executed at ring 3)", "TSD Set (RDTSC Can be executed only at ring 0)"},
         {"DE not Set" , "DE Set"},
         {"PSE not Set", "PSE Set"},
         {"PAE not Set", "PAE Set"},       /* 5 */
         {"MCE not Set", "MCE Set"},
         {"PGE not Set", "PGE Set"},
         {"PCE not Set (RDPMC Can execute only at Ring 0)", "PCE Set (RDPMC Can execute at Ring 3)"},       /* 8 */
         {"OSFXSR not Set", "OSFXSR Set"},
         {"OSXMMEXCPT not Set", "OSXMMEXCPT Set"} /* 10 */
       };

  /* PROGRAM VAR-S*/
  int ind;
  char msg_str[1024]; /* At most 1K char message */

  /**********************************/
  /* (1) INFO FROM asm/processor.h: */
  /**********************************/
  cpuid_level = current_cpu_data.cpuid_level;
  /*vendor_id = current_cpu_data.x86_vendor_id;*/
  /*model_id = current_cpu_data.x86_model_id;*/
  memcpy(vendor_id, current_cpu_data.x86_vendor_id, 16 * sizeof(char));
  memcpy(model_id, current_cpu_data.x86_model_id, 64 * sizeof(char));
  cache_size = current_cpu_data.x86_cache_size;

  /* Generate the message string */
  sprintf(msg_str,
          "\nDATA ACQUIRED FROM /ASM/PROCESSOR.H:%s",
        "\n\r====================================\n");
  /* Print to the current TTY */
  print_toTTY(msg_str);

  if (cpuid_level==-1)
  {
    /* Generate the message string */
    sprintf(msg_str,
            "\n\tCPUID INSTR-N NOT PRESENT (CPUID RETURNED %d \n\n",
            cpuid_level);
    print_toTTY(msg_str);
  }
  else
  {
    /* Generate the message string */
    sprintf(msg_str,"\tMAX. SUPPORTED CPUID LEVEL: %d",
            cpuid_level);
    print_toTTY(msg_str);
  }
  /* LKM insmod DOESN'T LIKE THESE:
  identify_cpu(&current_cpu_data);
  print_cpu_info(&current_cpu_data);*/
  sprintf(msg_str,"\tCPU VENDOR ID: %s",vendor_id);
  print_toTTY(msg_str);
  sprintf(msg_str,"\tCPU MODEL ID: %s",model_id);
  print_toTTY(msg_str);
  sprintf(msg_str,"\tCACHE SIZE: %d KB",cache_size);
  print_toTTY(msg_str);
  sprintf(msg_str,"\tCPU HAS TSC: %d ",cpu_has_tsc);
  print_toTTY(msg_str);


  /*****************************************************************/
  /* (2) INFO FROM INTEL CPUID INSTR-N: */
  /*****************************************************************/

  sprintf(msg_str,
          "\n\nDATA ACQUIRED FROM INTEL CPUID INSTRUCTION:%s",
          "\n\r===========================================\n");
  print_toTTY(msg_str);


  /* BASIC CPUID INFO */
  /*----------------- */

  /* EAX=0: */
  /*^^^^^^^ */
  sprintf(msg_str,
          "CPUID Level 0:%s",
      "\n\r--------------\n");
  print_toTTY(msg_str);

  pointer_to_char = (unsigned int *) vendor_id2;
  /* make it point to same loc. as vendor_id2 */
  eax_in = 0x00000000;
  MYcpuid(eax_in, &eax, &pointer_to_char[0], pointer_to_char+2, &pointer_to_char[1]);
  /* &pointer_to_char[1] = pointer_to_char+1 (unsigned int --> WORD ADDRESSED!!) */
  sprintf(msg_str,
          "\tEAX = %#010X ===> MAX supported EAX value for BASIC CPUID info = %d  \n"
          ,eax, eax);
  print_toTTY(msg_str);

  vendor_id2[12]='\0'; /* null char baby */
  sprintf(msg_str,"\tCPU VENDOR-ID: %s",vendor_id2);
  print_toTTY(msg_str);

  /* EAX=1: */
  /*^^^^^^^ */
  sprintf(msg_str,
          "\nCPUID Level 1:%s",
        "\n\r--------------\n");
  print_toTTY(msg_str);

  eax_in = 0x00000001;
  MYcpuid(eax_in, &eax, &ebx, &ecx, &edx);

  /* EAX */
  sprintf(msg_str,"\tEAX = %#010X (Processor Signature)\n", eax);
  print_toTTY(msg_str);
  /* EAX Dissection: */
  sprintf(msg_str,
          "\t\t Extended    Extended    Type    Family    Model     Stepping "
      "\n\r\t\t  Family      Model               Code     Number       ID    "
      "\n\r\t\t----------  ----------  ------  --------  --------  ----------"
      "\n\r\t\t %d%d%d%d%d%d%d%d     %d%d%d%d        %d%d      %d%d%d%d      %d%d%d%d       %d%d%d%d"
      "\n\r\t\t (Refer to Intel CPUID Manual Table 4 for corresponding Processor description)\n",
       testbit32(eax,27),testbit32(eax,26),testbit32(eax,25),testbit32(eax,24),
       testbit32(eax,23),testbit32(eax,22),testbit32(eax,21),testbit32(eax,20),
       testbit32(eax,19),testbit32(eax,18),testbit32(eax,17),testbit32(eax,16),
       testbit32(eax,13),testbit32(eax,12),
       testbit32(eax,11),testbit32(eax,10),testbit32(eax, 9),testbit32(eax, 8),
       testbit32(eax, 7),testbit32(eax, 6),testbit32(eax, 5),testbit32(eax, 4),
       testbit32(eax, 3),testbit32(eax, 2),testbit32(eax, 1),testbit32(eax, 0));
  print_toTTY(msg_str);
  /*Processor Type: */
  sprintf(msg_str,"\t\tProcessor Type: %s",
                  processor_type[testbit32(eax,13)*2 + testbit32(eax,12)]);
  print_toTTY(msg_str);
  /* Compute the Family & Model Numbers */
  extended_family_no = 0; family_code = 0;
  extended_model_no = 0; model_no = 0;
  stepping_id = 0;
  for (ind=27;ind>=20;ind--) /* compute ext. family # */
    extended_family_no = 2*extended_family_no + testbit32(eax,ind);
  for (ind=11;ind>=8;ind--) /* compute Family Code */
    family_code = 2*family_code + testbit32(eax,ind);
  for (ind=19;ind>=16;ind--) /* compute Ext. Model NO */
    extended_model_no = 2*extended_model_no + testbit32(eax,ind);
  for (ind=7;ind>=4;ind--) /* compute Model No */
    model_no = 2*model_no + testbit32(eax,ind);
  for (ind=3;ind>=0;ind--) /* compute Stepping ID */
    stepping_id = 2*stepping_id + testbit32(eax,ind);
  /*Processor Family: */
  if ((extended_family_no==0x00) && (family_code==0x06))
  {
    sprintf(msg_str,"\t\tProcessor Family: P6");
    print_toTTY(msg_str);
    /*Processor Model: */
    if ((extended_model_no==0x0) && (model_no==0x1))
      sprintf(msg_str,"\t\tProcessor Model: Pentium PRO");
    else if ((extended_model_no==0x00) && (model_no==0x3))
      sprintf(msg_str,"\t\tProcessor Model: PentiumII Model 3");
    else if ((extended_model_no==0x00) && (model_no==0x5))
      sprintf(msg_str,"\t\tProcessor Model: PentiumII Model 5");
    else if ((extended_model_no==0x00) && (model_no==0x6))
      sprintf(msg_str,"\t\tProcessor Model: Celeron");
    else if ((extended_model_no==0x00) && (model_no==0x7))
      sprintf(msg_str,"\t\tProcessor Model: PentiumIII Model 7");
    else if ((extended_model_no==0x00) && (model_no==0x8))
      sprintf(msg_str,"\t\tProcessor Model: PentiumIII Model 8");
    else if ((extended_model_no==0x00) && (model_no==0xA))
      sprintf(msg_str,"\t\tProcessor Model: PentiumIII Xeon Model A");
    else if ((extended_model_no==0x00) && (model_no==0xB))
      sprintf(msg_str,"\t\tProcessor Model: PentiumIII Xeon Model B");
    else
      sprintf(msg_str,"\t\tModel Unimplementd Yet... <cant>");
    print_toTTY(msg_str);
  }
  else if ((extended_family_no==0x00) && (family_code==0x0F))
  {
    sprintf(msg_str,"\t\tProcessor Family: Pentium 4 (Netburst)");
    print_toTTY(msg_str);
    /*Processor Model: */
    if ((extended_model_no==0x0) && (model_no==0x0))
      sprintf(msg_str,"\t\tProcessor Model: Pentium 4, 0.18U technology");
    else if ((extended_model_no==0x00) && (model_no==0x1))
      sprintf(msg_str,"\t\tProcessor Model: Pentium 4, 0.18U technology");
    else if ((extended_model_no==0x00) && (model_no==0x2))
      sprintf(msg_str,"\t\tProcessor Model: Pentium 4, 0.13U technology");
    else
      sprintf(msg_str,"\t\tModel Unimplementd Yet... <cant>");
    print_toTTY(msg_str);
  }
  else
  {
    sprintf(msg_str,"\t\tUnimplementd Yet... <cant>");
    print_toTTY(msg_str);
  }
  /* Use whole signature (including stepping: */
  sprintf(msg_str,
          "\t\t\nFOR CPU SIGNATURE: %#010X, POSSIBILITIES ARE: \n"
      "\n\r\t\t\t S-SPec    L2 Size    Core Speed    Bus Speed    Package "
      "\n\r\t\t\t--------  ---------  ------------  -----------  ---------",eax);
  print_toTTY(msg_str);
  if (eax==0x0F07)
      sprintf(msg_str,
          "\t\t\t SL4QD      256K       1.30GHz       400MHz     31.0mm OOI(423 pin)\n\r"
          "\t\t\t SL4SF      256K       1.30GHz       400MHz     31.0mm OOI(423 pin)\n\r"
          "\t\t\t SL4SG      256K       1.40GHz       400MHz     31.0mm OOI(423 pin)\n\r"
          "\t\t\t SL4SC      256K       1.40GHz       400MHz     31.0mm OOI(423 pin)\n\r"
          "\t\t\t SL4SH      256K       1.50GHz       400MHz     31.0mm OOI(423 pin)\n\r"
          "\t\t\t SL4TY      256K       1.50GHz       400MHz     31.0mm OOI(423 pin)");
  else if (eax==0x0F24)
      sprintf(msg_str,
          "\t\t\t SL5YS      512K       2.20GHz       400MHz     31.0mm FC\n\r"
          "\t\t\t SL68S      512K       2.20GHz       400MHz     31.0mm FC\n\r"
          "\t\t\t ........................................................");
          /* there is a whole bunch others but i didn't implement*/
  else
    sprintf(msg_str,"\t\t\tI HAVE NOT implementd Yet...Refer to P4 Specification Update... <cant>");
  print_toTTY(msg_str);
  print_toTTY("\t\t(Refer to Intel P4 SPECIFICATION UPDATE for COMPLETE SET)\n");

  /* EBX */
  sprintf(msg_str,"\tEBX = %#010X \n", ebx);
  print_toTTY(msg_str);
  /* EbX Dissection: */
  sprintf(msg_str,
          "\t\t Physical    # of Logical     CLFLUSH      Brand  "
      "\n\r\t\t APIC ID      Processors     line size      ID    "
      "\n\r\t\t----------  --------------  -----------  ---------"
      "\n\r\t\t %d%d%d%d%d%d%d%d     %d%d%d%d%d%d%d%d       %d%d%d%d%d%d%d%d     %d%d%d%d%d%d%d%d\n",
       testbit32(ebx,31),testbit32(ebx,30),testbit32(ebx,29),testbit32(ebx,28),
       testbit32(ebx,27),testbit32(ebx,26),testbit32(ebx,25),testbit32(ebx,24),
       testbit32(ebx,23),testbit32(ebx,22),testbit32(ebx,21),testbit32(ebx,20),
       testbit32(ebx,19),testbit32(ebx,18),testbit32(ebx,17),testbit32(ebx,16),
       testbit32(ebx,15),testbit32(ebx,14),testbit32(ebx,13),testbit32(ebx,12),
       testbit32(ebx,11),testbit32(ebx,10),testbit32(ebx, 9),testbit32(ebx, 8),
       testbit32(ebx, 7),testbit32(ebx, 6),testbit32(ebx, 5),testbit32(ebx, 4),
       testbit32(ebx, 3),testbit32(ebx, 2),testbit32(ebx, 1),testbit32(ebx, 0));
  print_toTTY(msg_str);
  /* Define Processor From Brand ID: */
  sprintf(msg_str,"\t\tProcessor Brand: %s",brand_table[(ebx & 0x000000FF)]);
  print_toTTY(msg_str);

  /* ECX */
  sprintf(msg_str,"\t\nECX = %#010X ", ecx);
  print_toTTY(msg_str);
  if (testbit32(ecx,10))
    sprintf(msg_str,"\n\t\t CID flag set");
  else
    sprintf(msg_str,"\n\t\t CID flag NOT set");
  print_toTTY(msg_str);


  /* EDX */
  sprintf(msg_str,"\t\nEDX = %#010X \n", edx);
  print_toTTY(msg_str);
  /* Print individual FLAG information */
  print_toTTY("\t\tPROCESSOR FLAGS:\n\r"
              "\t\t^^^^^^^^^^^^^^^^");
  /* Below Mostly Copied from Intel's CPUID3.c */
  if (edx & FPU_FLAG) print_toTTY("\t\tCPU contains an on-chip FPU");
  if (edx & VME_FLAG) print_toTTY("\t\tCPU supports Virtual Mode Extensions");
  if (edx & DE_FLAG) print_toTTY("\t\tCPU supports the Debugging Extensions");
  if (edx & PSE_FLAG) print_toTTY("\t\tCPU supports Page Size Extensions");
  if (edx & TSC_FLAG) print_toTTY("\t\tCPU supports Time Stamp Counter ==========> OUR STUFF!!");
  if (edx & MSR_FLAG) print_toTTY("\t\tCPU supports Model Specific Registers ==========> OUR STUFF!!");
  if (edx & PAE_FLAG) print_toTTY("\t\tCPU supports Physical Address Extension");
  if (edx & MCE_FLAG) print_toTTY("\t\tCPU supports Machine Check Exceptions");
  if (edx & CX8_FLAG) print_toTTY("\t\tCPU supports the CMPXCHG8B instruction");
  if (edx & APIC_FLAG) print_toTTY("\t\tCPU contains an on-chip APIC");
  if (edx & SEP_FLAG) print_toTTY("\t\tCPU supports the Fast System Call");
  if (edx & MTRR_FLAG) print_toTTY("\t\tCPU supports the Memory Type Range Registers");
  if (edx & PGE_FLAG) print_toTTY("\t\tCPU supports Page Global Enable");
  if (edx & MCA_FLAG) print_toTTY("\t\tCPU supports the Machine Check Architecture");
  if (edx & CMOV_FLAG) print_toTTY("\t\tCPU supports the Conditional Move Instruction");
  if (edx & PAT_FLAG) print_toTTY("\t\tCPU supports the Page Attribute Table");
  if (edx & PSE36_FLAG) print_toTTY("\t\tCPU supports 36-bit Page Size Extension");
  if (edx & PSNUM_FLAG) print_toTTY("\t\tCPU supports CPU serial number");
  if (edx & CLFLUSH_FLAG) print_toTTY("\t\tCPU supports the CLFLUSH instruction");
  if (edx & DTS_FLAG) print_toTTY("\t\tCPU supports the Debug Trace Store feature ========> OUR STUFF!!");
  if (edx & ACPI_FLAG) print_toTTY("\t\tCPU supports ACPI registers in MSR space");
  if (edx & MMX_FLAG) print_toTTY("\t\tCPU supports Intel Architecture MMX(TM) technology");
  if (edx & FXSR_FLAG) print_toTTY("\t\tCPU supports the Fast floating point save and restore");
  if (edx & SSE_FLAG) print_toTTY("\t\tCPU supports the Streaming SIMD extensions to the Intel Architecture");
  if (edx & SSE2_FLAG) print_toTTY("\t\tCPU supports the Streaming SIMD extensions 2 instructions");
  if (edx & SS_FLAG) print_toTTY("\t\tCPU supports Self-Snoop");
  if ((edx & HTT_FLAG) && (((ebx >> 16) & 0x0FF) > 1)) print_toTTY("\t\tCPU supports Hyper-Threading Technology");
  if (edx & TM_FLAG) print_toTTY("\t\tCPU supports the Thermal Monitor ==========> ??");
  if (edx & SBF_FLAG) print_toTTY("\t\tCPU supports the Signal Break on FERR Feature");

  /* EAX=2: */
  /*^^^^^^^ */
  sprintf(msg_str,
          "CPUID Level 2:%s",
      "\n\r--------------\n");
  print_toTTY(msg_str);
  sprintf(msg_str,"\t<CACHE AND TLB CHARACTERISTICS DESCRIPTORS>\n\r"
                  "\t(NOT IMPLEMENTED)");
  print_toTTY(msg_str);

  /* EXTENDED CPUID INFO */
  /*-------------------- */

  sprintf(msg_str,
        "\nEXTENDED CPUID Levels:%s",
      "\n\r----------------------\n");
  print_toTTY(msg_str);
  sprintf(msg_str,"\t<ONLY PRINTS THE BRAND ID STRING --> we already have from asm/processor.h>\n\r"
                  "\t(NOT IMPLEMENTED)");
  print_toTTY(msg_str);


  /*****************************************************************/
  /* (3) INFO FROM INTEL CONTROL REGISTERS (ONLY CR4 IMPLEMENTED): */
  /*****************************************************************/

  sprintf(msg_str,
          "\n\nDATA ACQUIRED FROM INTEL CONTROL REGISTERS:%s",
          "\n\r===========================================\n");
  print_toTTY(msg_str);

  /* Although CR4 is in asm/processor.h --> mmu_cr4_features,
   *I cannot access is as i386_ksyms.c, doesn't export it!!*/
  /*sprintf(msg_str,"\tCR4: %#9X ", mmu_cr4_features);
  print_toTTY(msg_str);*/

  /* CR4 */
  /*=====*/
  sprintf(msg_str,
          "CR4 :%s",
      "\n\r-----\n");
  print_toTTY(msg_str);
  rdCR4(CR4contents);
  sprintf(msg_str,"\tCR4 REG contents: %#010X\n", CR4contents);
  print_toTTY(msg_str);

  for(ind=0;ind<=10;ind++)
  {
    sprintf(msg_str,"\tCR4 bit[%d] = %d --> %s ",
                    ind, testbit32(CR4contents,ind),
                    CR4_bitmessages[ind][testbit32(CR4contents,ind)]);
    print_toTTY(msg_str);
  }

  print_toTTY("\n\t\t\t\tDONE!");


  printk("Syscall getCPUinfo Says hello to Console\n");

  return 0;
}

/*this function is called when the module is
 *loaded (initialization)*/
int init_module()
{
  /*store original reference of syscall table for entry 222*/
  old_reference222 = (void *) sys_call_table[MY_SYSCALL_NO1];

  /*Add our syscall to syscall table [222] */
  sys_call_table[MY_SYSCALL_NO1]=(void *) sys_getCPUinfo;

  print_toTTY("\nmodule LKM_CPUinfo inserted\n");
  printk("hello from init of LKM_CPUinfo to console\n");
  return 0;
}

/*this function is called when the module is unloaded*/
 void cleanup_module()
 {
   /*make syscall_table[222] point to original reference when our module
    *is unloaded*/
   sys_call_table[MY_SYSCALL_NO1]=old_reference222;

   print_toTTY("\nmodule LKM_CPUinfo removed\n");
   printk("BYE from cleanup of LKM_CPUinfo to console\n");
 }



