/* MY LINUX CODE TO DEFINE Performance Reading SYSCALLs:
 * select_events, get_event_counts
 * 
 * lowpowered by: canturk isci / canturk@princeton.edu / YOHOO
 */

/* I AM USING SYSCALL # 222 and UP FOR MY SYSCALLs AS  ASM/UNISTD.H GOES
  *UP TO 221 */

/*
COMPILE WITH:
gcc -DMODULE -D__KERNEL__ -DLINUX -Wall -I/lib/modules/`uname -r`/build/include
-c LKM_PerformanceReader.c -o LKM_PerformanceReader.o
(You can get rid of some of the compile -D directives by defining them in code, but who cares?)
*/

/* Standard in kernel modules */
#include <linux/kernel.h>
#include <linux/module.h>

/* For syscalls*/
#include <sys/syscall.h>
#include <linux/sys.h>

/* For tty */
#include <linux/sched.h> /* For current */
#include <linux/tty.h> /* For the tty declarations */

/* Deal with CONFIG_MODVERSIONS */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <stdio.h>

#include <asm/msr.h> /* for the rdmsr, wrmsr stuff */

#include <asm/uaccess.h> /* for copy to/from user */

/* DEFINE THE NUMBERs I CHOSE FOR MY SYSCALLs */
/* (May be made user enterable on command line while insmod */
#define MY_SYSCALL_NOselect_events 223 /* 222 is for getCPUinfo, so I go on from 223 */
#define MY_SYSCALL_NOreset_event_counter 224 /* 222 is for getCPUinfo, so I go on from 223 */
#define MY_SYSCALL_NOstart_event_counter 225 /* 222 is for getCPUinfo, so I go on from 223 */
#define MY_SYSCALL_NOstop_event_counter 226 /* 222 is for getCPUinfo, so I go on from 223 */
#define MY_SYSCALL_NOget_event_counts 227 /* 222 is for getCPUinfo, so I go on from 223 */
#define MY_SYSCALL_NOset_replay_MSRs 228

/* INCLUDE SYSTEM CALL TABLE to change it*/
extern void *sys_call_table[];

/* MY syscalls */
asmlinkage static int sys_select_events(void * ESCRFieldsFromUser, void * CCCRFieldsFromUser);
asmlinkage static int sys_reset_event_counter(unsigned int counter);
asmlinkage static int sys_start_event_counter(unsigned int CCCR);
asmlinkage static int sys_stop_event_counter(unsigned int CCCR);
asmlinkage static int sys_get_event_counts(unsigned long long * datatouser);
asmlinkage static int sys_set_replay_MSRs(unsigned int MSR, unsigned long long setbits);

/* Old Syscall table references for the new syscall numbers */
void *old_reference223;
void *old_reference224;
void *old_reference225;
void *old_reference226;
void *old_reference227;
void *old_reference228;

/* MY MACRO TO TEST A BIT IN A 64 bit DATA */
/*return 0 if bit at bitnum = 0; 1 if bit = 1 */
#define testbit64(testdata,bitnum)\
        ((((0x0000000000000000 | (1 << bitnum)) & testdata)==0) ? 0 : 1)

/* MACRO TO CONCAT 2 32 BIT DATA TO MAKE A 64 BIT DATA */
#define concat32_32(Hi32bits,Lo32bits) /* Make a 64 bit result */\
        ((((unsigned long long int) Hi32bits) << 32) | (unsigned long long int) Lo32bits)

/* READ TIME STAMP COUNTER MACRO, SAME AS RUSS */
#define read_tsc(TSCvalue) {\
        unsigned int __low32, __hi32;\
        rdtsc(__low32,__hi32); /*from msr.h */\
        TSCvalue = concat32_32(__hi32,__low32);\
}

/* Read MSR MACRO: Reads 64 bit MSR value of MSR[MSR_address] */
#define read_msr(MSR_address,MSR_value64){\
        unsigned int __low32, __hi32;\
        rdmsr(MSR_address,__low32,__hi32); /*from msr.h */\
        MSR_value64 = concat32_32(__hi32,__low32);\
}


/* Write MSR MACRO: writes 64 bit MSR value to MSR[MSR_address] */
#define write_msr(MSR_address,MSR_value64){\
        unsigned int __low32, __hi32;\
        __low32 = (unsigned int) (MSR_value64 & 0x00000000FFFFFFFF);\
        __hi32  = (unsigned int) ((MSR_value64 >> 32) & 0x00000000FFFFFFFF);/*AND to be safe*/\
        wrmsr(MSR_address,__low32,__hi32); /*from msr.h */\
}

/* MSR ADDRESSES:*/
#define IA32_TIME_STAMP_COUNTER 0x0010 /* MSR Address of TSC */
#define IA32_MISC_ENABLE        0x01A0 /* MSR TO ENABLE/DISABLE thermal Monitoring,
                                         perfoamnace monitorung, PEBS and etc.  */

struct ESCR_struct { unsigned int ESCR;       unsigned int Event_Select;
                     unsigned int Event_Mask; unsigned int Tag_Value;
                     unsigned int Tag_En;     unsigned int OS;
                     unsigned int USR;
                   } ESCR_Fields;

struct CCCR_struct { unsigned int CCCR;        unsigned int OVF;
                     unsigned int Cascade;     unsigned int OVF_PMI;
                     unsigned int Force_OVF;   unsigned int Edge;
                     unsigned int Threshold;   unsigned int Complement;
                     unsigned int Compare;     unsigned int Reserved_17_16;
                     unsigned int ESCR_Select; unsigned int EN;
                   } CCCR_Fields;

char msg_str[128]; /* At most 128B char message */

/* Function to print to the tty the current task uses */
/*Apparently, only prints up to 4096 bytes, ignores the rest*
 *tty->driver.flush_buffer can be used to flush it, but it  *
 *doesn't make it write the flushed portion to the tty device*/
void print_toTTY(char *str)
  {
    struct tty_struct *my_tty;

    /* The tty for the current task */
    my_tty = current->tty;

    /* If my_tty is NULL, it means that the current task
    * has no tty you can print to */

    if (my_tty != NULL) {

      /* my_tty->driver is a struct which holds the tty's
       * functions, one of which (write) is used to
       * write strings to the tty. It can be used to take
       * a string either from the user's memory segment
       * or the kernel's memory segment.
       *
       * The function's first parameter is the tty to
       * write to, because the  same function would
       * normally be used for all tty's of a certain type.
       * The second parameter controls whether the
       * function receives a string from kernel memory
       * (false, 0) or from user memory (true, non zero).
       * The third parameter is a pointer to a string,
       * and the fourth parameter is the length of
       * the string.
       */
      (*(my_tty->driver).write)(
         my_tty, /* The tty itself */
         0, /* We don't take the string from user space */
         str, /* String */
         strlen(str));  /* Length */
      (*(my_tty->driver).write)(
         my_tty,
         0,
         "\015\012",
         2);
    }
  }


int construct_ESCR(struct ESCR_struct * fields)
{
  unsigned long long ESCR_value;
  /* Read Current ESCR first */
  read_msr(fields->ESCR, ESCR_value);
  /* Print Original Value */
  sprintf(msg_str,"\tOriginal ESCR[%#04X]:      %#018X", fields->ESCR, ESCR_value);
  print_toTTY(msg_str);

  /* clear all the fields to be written first (untouch others)*/
  ESCR_value = ESCR_value & 0xFFFFFFFF80000003;
  /* Set the field bits: */
  ESCR_value = ESCR_value | ((fields->Event_Select) << 25)
                          | ((fields->Event_Mask) << 9)
                          | ((fields->Tag_Value) << 5)
                          | ((fields->Tag_En) << 4)
                          | ((fields->OS) << 3)
                          | ((fields->USR) << 2);
  /* Print Written ESCR Value */
  sprintf(msg_str,"\tESCR[%#04X] Configuration: %#018X", fields->ESCR, ESCR_value);
  print_toTTY(msg_str);
  /* Write the ESCR configuration to ESCR */
  write_msr(fields->ESCR, ESCR_value);
  /* Read back to verify */
  read_msr(fields->ESCR, ESCR_value);
  sprintf(msg_str,"\tVerified ESCR[%#04X]:      %#018X\n", fields->ESCR, ESCR_value);
  print_toTTY(msg_str);

  return 0;
}

int construct_CCCR(struct CCCR_struct * fields)
{
  unsigned long long CCCR_value;
  /* Read Current CCCR first */
  read_msr(fields->CCCR, CCCR_value);
  /* Print Original Value */
  sprintf(msg_str,"\tOriginal CCCR[%#04X]:      %#018X", fields->CCCR, CCCR_value);
  print_toTTY(msg_str);

  /* clear all the fields to be written first (untouch others)*/
  CCCR_value = CCCR_value & 0xFFFFFFFF38000FFF;
  /* Set the field bits: */
  CCCR_value = CCCR_value | ((fields->OVF) << 31)
                          | ((fields->Cascade) << 30)
                          | ((fields->OVF_PMI) << 26)
                          | ((fields->Force_OVF) << 25)
                          | ((fields->Edge) << 24)
                          | ((fields->Threshold) << 20)
                          | ((fields->Complement) << 19)
                          | ((fields->Compare) << 18)
                          | ((fields->Reserved_17_16) << 16)
                          | ((fields->ESCR_Select) << 13)
                          | ((fields->EN) << 12);

  /* Print Written CCCR Value */
  sprintf(msg_str,"\tCCCR[%#04X] Configuration: %#018X", fields->CCCR, CCCR_value);
  print_toTTY(msg_str);
  /* Write the CCCR configuration to CCCR */
  write_msr(fields->CCCR, CCCR_value);
  /* Read back and verify */
  read_msr(fields->CCCR, CCCR_value);
  sprintf(msg_str,"\tVerified CCCR[%#04X]:      %#018X", fields->CCCR, CCCR_value);
  print_toTTY(msg_str);

  return 0;
}

/* DEFINITION OF MY SYSCALL #223*/
/* Read the ESCR data & CCCR data from user space and write to ESCR & CCCR reg-s */
/*If necessary define types to this syscall( int type), for events that require >1 ESCR etc.
for each type copy_from_user' lar deisik bi struct alir olur biter */
asmlinkage static int sys_select_events(void * ESCRFieldsFromUser, void * CCCRFieldsFromUser)
{
  /*ESCR Fields: */
  if(copy_from_user(&ESCR_Fields, ESCRFieldsFromUser, sizeof(struct ESCR_struct)))
    print_toTTY("\n\tCOULD NOT COPY ESCR FIELDS FROM USER CORRECTLY");
  /*CCCR Fields: */
  if(copy_from_user(&CCCR_Fields, CCCRFieldsFromUser, sizeof(struct CCCR_struct)))
    print_toTTY("\n\tCOULD NOT COPY CCCR FIELDS FROM USER CORRECTLY");
  /* Construct The Registers */
  if(ESCR_Fields.ESCR) /* Nonzero ESCR */
  {
    sprintf(msg_str,"\nSetting ESCR & CCCR for Counter %d:\n", (CCCR_Fields.CCCR - 0x0360));
    print_toTTY(msg_str);
    construct_ESCR(&ESCR_Fields);
    construct_CCCR(&CCCR_Fields);
  }
  else
  {
    sprintf(msg_str,"\nCounter %d UNUSED", (CCCR_Fields.CCCR - 0x0360));
    print_toTTY(msg_str);
    return -1;
  }

  return 0;
}

/* DEFINITION OF MY SYSCALL #224*/
asmlinkage static int sys_reset_event_counter(unsigned int counter)
{
  unsigned long long counter_value; /* Counter value */

  sprintf(msg_str,"\nResetting Counter %d:\n", (counter - 0x0300));
  print_toTTY(msg_str);
  /* Print Original Value */
  read_msr(counter, counter_value);
  sprintf(msg_str,"\tOriginal counter[%#04X]:      %#018X", counter, counter_value);
  print_toTTY(msg_str);
  /* clear all the unreserved 40 bits*/
  counter_value = counter_value & 0xFFFFFF0000000000;
  write_msr(counter, counter_value);
  /* Read back and verify */
  read_msr(counter, counter_value);
  sprintf(msg_str,"\tVerified counter[%#04X]:      %#018X", counter, counter_value);
  print_toTTY(msg_str);

  return 0;
}

/* DEFINITION OF MY SYSCALL #225*/
asmlinkage static int sys_start_event_counter(unsigned int CCCR)
{
  unsigned long long CCCR_value; /* Counter value */

  sprintf(msg_str,"\nEnabling Counter %d:\n", (CCCR - 0x0360));
  print_toTTY(msg_str);
  /* Print Original Value */
  read_msr(CCCR, CCCR_value);
  sprintf(msg_str,"\tOriginal CCCR[%#04X]:      %#018X", CCCR, CCCR_value);
  print_toTTY(msg_str);
  /* SET ENABLE BIT (#12) */
  CCCR_value = CCCR_value | (0x01 << 12);
  write_msr(CCCR, CCCR_value);
  /* Read back and verify */
  read_msr(CCCR, CCCR_value);
  sprintf(msg_str,"\tVerified CCCR[%#04X]:      %#018X", CCCR, CCCR_value);
  print_toTTY(msg_str);

  return 0;
}

/* DEFINITION OF MY SYSCALL #226*/
asmlinkage static int sys_stop_event_counter(unsigned int CCCR)
{
  unsigned long long CCCR_value; /* Counter value */

  sprintf(msg_str,"\nStopping Counter %d:\n", (CCCR - 0x0360));
  print_toTTY(msg_str);
  /* Print Original Value */
  read_msr(CCCR, CCCR_value);
  sprintf(msg_str,"\tOriginal CCCR[%#04X]:      %#018X", CCCR, CCCR_value);
  print_toTTY(msg_str);
  /* CLEAR ENABLE BIT (#12) */
  CCCR_value = CCCR_value & (0xFFFFFFFFFFFFEFFF);
  write_msr(CCCR, CCCR_value);
  /* Read back and verify */
  read_msr(CCCR, CCCR_value);
  sprintf(msg_str,"\tVerified CCCR[%#04X]:      %#018X", CCCR, CCCR_value);
  print_toTTY(msg_str);

  return 0;
}

/* DEFINITION OF MY SYSCALL #227*/
/* copies the content of 18 counters and TSC */
asmlinkage static int sys_get_event_counts(unsigned long long * datatouser)
{
  unsigned int counters[18] = {0x0300, 0x0301, 0x0302, 0x0303,/*0,1,2,3*/
                               0x0304, 0x0305, 0x0306, 0x0307,/*4,5,6,7*/
                               0x0308, 0x0309, 0x030A, 0x030B,/*8,9,10,11*/
                               0x030C, 0x030D, 0x030E, 0x030F,/*12,13,14,15*/
                               0x0310, 0x0311}; /*16,17*/

  unsigned long long countervalues[19];
  /* 0-17: for counters; 18 for TSC*/
  int ind;

  /* Read COunters and TSC */
  for (ind=0; ind<=17; ind++)
  {
    read_msr(counters[ind], countervalues[ind]);
    /* MASK BITS ABOVE 40 */
    countervalues[ind] = countervalues[ind] & 0x000000FFFFFFFFFF;
  }
  read_tsc(countervalues[18]);
  /* COPY DATA to user, does VERIFY_WRITE internally */
  if(copy_to_user(datatouser, &countervalues, 19 * sizeof(unsigned long long)))
    print_toTTY("\n\tCOULD NOT COPY COUNTER DATA TO USER CORRECTLY");

  return 0;
}

/* DEFINITION OF MY SYSCALL #228*/
/* Give the setbits so that only the bits to be set are 1, it's OR ed with the original */
asmlinkage static int sys_set_replay_MSRs(unsigned int MSR, unsigned long long setbits)
{
  unsigned long long MSR_value; /* original MSR value */

  sprintf(msg_str,"\nSetting MSR[%#04X]:\n", (MSR));
  print_toTTY(msg_str);
  /* check if MSR_PEBS_MATRIX_VERT or IA32_PEBS_ENABLE for security*/
  if (MSR < 0x3F1 || MSR > 0x3F2)
  {
    print_toTTY("\n\tCOULD NOT COPY COUNTER DATA TO USER CORRECTLY");
    return -1;
  }
  /* Print Original Value */
  read_msr(MSR, MSR_value);
  sprintf(msg_str,"\tOriginal MSR[%#04X]:      %#018X", MSR, MSR_value);
  print_toTTY(msg_str);
  /* OR the current value with the supplied setbits */
  MSR_value = setbits;/*MSR_value | setbits;*//*I REMOVED THE OR FOR NOW*/
  write_msr(MSR, MSR_value);
  /* Read back and verify */
  read_msr(MSR, MSR_value);
  sprintf(msg_str,"\tVerified MSR[%#04X]:      %#018X", MSR, MSR_value);
  print_toTTY(msg_str);

  return 0;
}



/*this function is called when the module is
 *loaded (initialization)*/
int init_module()
{
  /*My Beautiful 3D structure to print info about IA32_MISC_ENABLE bits */
  /*Each row is for one bit of MISC_ENABLE, starting from bit0 */
  /*Prints left message if bit not set, right one if bit set */
  char IA32_MISC_EN_bitmessages[][2][128] = {
         {"Fast Strings Disabled", "Fast Strings Enabled"},       /* 0 */
         {"Reserved", "Reserved"},       /* 1 */
         {"Fopcode Compatibility Mode Disabled", "Fopcode Compatibility Mode Enabled"},
         {"Thermal clock modulation disabled" , "Thermal clock modulation Enabled"},
         {"Split Lock Enabled", "Split Lock Disabled"},
         {"Reserved", "Reserved"},       /* 5 */
         {"L3 Cache Enabled", "L3 Cache Disabled"},
         {"Performance Monitoring Disabled", "Performance Monitoring Available"},
         {"Suppress Lock Disabled", "Suppress Lock Enabled"},       /* 8 */
         {"Prefetch Queue Enabled", "Prefetch Queue Disabled"},
         {"FERR Disabled", "FERR Enabled"}, /* 10 */
         {"Branch Trace Storage Available", "Branch Trace Storage Unavailable"}, /* 11 */
         {"PEBS Available", "PEBS UNavailable"} /* 12 */
       };
  unsigned long long IA32_MISC_EN_value;
/*  char msg_str[128];*/ /* At most 128B char message */
  int ind;

  /*store original references of syscall table for entry 223-227*/
  old_reference223 = (void *) sys_call_table[MY_SYSCALL_NOselect_events];
  old_reference224 = (void *) sys_call_table[MY_SYSCALL_NOreset_event_counter];
  old_reference225 = (void *) sys_call_table[MY_SYSCALL_NOstart_event_counter];
  old_reference226 = (void *) sys_call_table[MY_SYSCALL_NOstop_event_counter];
  old_reference227 = (void *) sys_call_table[MY_SYSCALL_NOget_event_counts];
  old_reference228 = (void *) sys_call_table[MY_SYSCALL_NOset_replay_MSRs];

  /*Add our syscall to syscall table [223-228] */
  sys_call_table[MY_SYSCALL_NOselect_events]=(void *) sys_select_events;
  sys_call_table[MY_SYSCALL_NOreset_event_counter]=(void *) sys_reset_event_counter;
  sys_call_table[MY_SYSCALL_NOstart_event_counter]=(void *) sys_start_event_counter;
  sys_call_table[MY_SYSCALL_NOstop_event_counter]=(void *) sys_stop_event_counter;
  sys_call_table[MY_SYSCALL_NOget_event_counts]=(void *) sys_get_event_counts;
  sys_call_table[MY_SYSCALL_NOset_replay_MSRs]=(void *) sys_set_replay_MSRs;

  /* Read IA32_MISC_ENABLE MSR to see availability of pErf. cntrs */
  /*It might better to do this in getCPUinfo syscall, but I ran out of 4K buffer*/
  read_msr(IA32_MISC_ENABLE,IA32_MISC_EN_value);
  sprintf(msg_str,
          "\n\nIA32_MISC_ENABLE MSR Contents:"
          "\n\r==============================\n");
  print_toTTY(msg_str);
  sprintf(msg_str,"%#018X\n", IA32_MISC_EN_value);
  print_toTTY(msg_str);
  for(ind=0;ind<=12;ind++)
  {
    sprintf(msg_str,"\tIA32_MISC_ENABLE bit[%d] = %d --> %s ",
                    ind, testbit64(IA32_MISC_EN_value,ind),
                    IA32_MISC_EN_bitmessages[ind][testbit64(IA32_MISC_EN_value,ind)]);
    print_toTTY(msg_str);
  }

  print_toTTY("\nmodule LKM_PerformanceReader inserted\n");
  printk("hello from init of LKM_PerformanceReader to console\n");
  return 0;
}

/*this function is called when the module is unloaded*/
 void cleanup_module()
 {
   /*make syscall_table[223] point to original reference when our module
    *is unloaded*/
   sys_call_table[MY_SYSCALL_NOselect_events]=old_reference223;
   sys_call_table[MY_SYSCALL_NOreset_event_counter]=old_reference224;
   sys_call_table[MY_SYSCALL_NOstart_event_counter]=old_reference225;
   sys_call_table[MY_SYSCALL_NOstop_event_counter]=old_reference226;
   sys_call_table[MY_SYSCALL_NOget_event_counts]=old_reference227;
   sys_call_table[MY_SYSCALL_NOset_replay_MSRs]=old_reference228;

   print_toTTY("\nmodule LKM_PerformanceReader removed\n");
   printk("BYE from cleanup of LKM_PerformanceReader to console\n");
 }



