
 /*COMPILE WITH:
    make -f Makefile_PowerServer
 */
/* Includes Performance counter reading interface and sending results over ether
 * USe as: PowerServer <portnum> (in conjunction with PowerClient)
 * only need to change the PerformanceReader loop
 * 
 * lowpowered by: canturk isci / canturk@princeton.edu / DRINK DEW!!
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* socket stuff*/
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <unistd.h> /* SLEEP AND USLEEP HERE */

#include <string.h>

#include "UserLevel_PerformanceReader.h" /*syscall stuff */
#include "PowerServer_DefineEvents.c" /*Perf Counter Stuff*/

/* ALL 24 COUNTED EVENTS */
enum {   eventcount_IOQallocation =0, eventcount_FSBdataactivity,  eventcount_BSQcacheref,
         eventcount_ITLBref,          eventcount_branchretired,    eventcount_BPUfetchreq,
         eventcount_LDportreplay,     eventcount_STportreplay,     eventcount_Frontend,
         eventcount_UOPtype,          eventcount_MOBloadreplay,    eventcount_machineclear,
         eventcount_UOPqueuewrites07, eventcount_PackedSPuop,      eventcount_PackedDPuop,
         eventcount_scalarSPuop,      eventcount_scalarDPuop,      eventcount_64bitMMXuop,
         eventcount_128bitMMXuop,     eventcount_x87FPuop,         eventcount_x87SIMDmovesuop,
         eventcount_TCdelivermode,    eventcount_UOPqueuewrites04, eventcount_UOPSretired, /*23*/
         NUM_EVENTS /*24*/
};

/* server realated: */
#define MAX_CHAR 255
#define ERROR -1
#define BACKLOG 5

static int socket_write(int sock, char *buf, int len)
{
  return write(sock, buf, len); /* writes to filedes, and blocks until writes if O_NONBLOCK not set*/
}

void PerformanceReader (int sock)
/* ALL COUNTER READING STUFF IS DONE HERE */
{
  int i;         /* loop counter */
  unsigned long long CounterValues[19];
  /* For ALL these arrays, 1st 18 are counter values, 19th is the TSC */
  /* ALL TIMESTAMPS */
  double timestamp[5] = {0.0};
  /* ALL 24 COUNTED EVENTS */
  double eventcounts[NUM_EVENTS] = {0.0};
  const static unsigned int dt = 100*1000; /* sampling period (delta t) for serial data [Useconds] */
  /* FOR SOCKET WRITING:*/
  char *pos;

  /**********************************************************/
  /* SELECT EVENTS AND RESET THE COUNTERS  */
  /**********************************************************/
  /* (2) CLEAR COUNTERS */
  for (i=0;i<=17;i++)
  {
    reset_event_counter(counters[i]);
  }

  /* (1) DEFINE EVENTS TO COUNT w/o ROTATION */
  define_events(); /* From DefineEvents.c */

  /***********************/
  /* (3) START COUNTERS */
  /***********************/
  /* ENABLED DIRECTLY WHEN EVENT IS SELECTED */
  /* NO NEED FOR SEPARATE START_EVENT_COUNTER*/


  /*--------------------------------------------------------------*/
  /* Main COunter Loop */
  /*--------------------------------------------------------------*/

  while(1)
  {
    char buffer[800]; /* BUFFER TO WRITE TO SOCKET */

    /* RESET UNROTATED EVENTS */
    reset_event_counter(counters[1]);
    reset_event_counter(counters[3]);
    reset_event_counter(counters[4]);
    reset_event_counter(counters[5]);
    reset_event_counter(counters[6]);
    reset_event_counter(counters[9]);
    reset_event_counter(counters[11]);
    reset_event_counter(counters[13]);
    reset_event_counter(counters[14]);
    reset_event_counter(counters[15]);
    /* GET TIMESTAMP FOR NOW */
    get_event_counts(CounterValues);
    timestamp[0] =(double) CounterValues[18];

    /* ROTATION 1*/
    /*************/
    define_events_rotate1();
    /* reset rotation 1 only events*/
    reset_event_counter(counters[8]);
    reset_event_counter(counters[10]);
    /* reset rotation 1&2 only events*/
    reset_event_counter(counters[0]);
    reset_event_counter(counters[2]);
    reset_event_counter(counters[12]);
    usleep(dt);
    get_event_counts(CounterValues);
    /* counts for rotation 1 only events*/
    eventcounts[eventcount_PackedSPuop] = (double) CounterValues[8];
    eventcounts[eventcount_PackedDPuop] = (double) CounterValues[10];
    /* timestamp for the end of rotation 1 */
    timestamp[1] =(double) CounterValues[18];

    /* ROTATION 2*/
    /*************/
    define_events_rotate2();
    /* reset rotation 2 only events*/
    reset_event_counter(counters[8]);
    reset_event_counter(counters[10]);
    usleep(dt);
    get_event_counts(CounterValues);
    /* counts for rotation 2 only events*/
    eventcounts[eventcount_scalarSPuop] = (double) CounterValues[8];
    eventcounts[eventcount_scalarDPuop] = (double) CounterValues[10];
    /* counts for rotation 1&2 only events*/
    eventcounts[eventcount_IOQallocation] = (double) CounterValues[0];
    eventcounts[eventcount_BPUfetchreq]   = (double) CounterValues[2];
    eventcounts[eventcount_branchretired] = (double) CounterValues[12];
    /* timestamp for the end of rotation 2 */
    timestamp[2] =(double) CounterValues[18];

    /* ROTATION 3*/
    /*************/
    define_events_rotate3();
    /* reset rotation 3 only events*/
    reset_event_counter(counters[8]);
    reset_event_counter(counters[10]);
    /* reset rotation 3&4 only events*/
    reset_event_counter(counters[0]);
    reset_event_counter(counters[2]);
    reset_event_counter(counters[12]);
    usleep(dt);
    get_event_counts(CounterValues);
    /* counts for rotation 3 only events*/
    eventcounts[eventcount_64bitMMXuop]  = (double) CounterValues[8];
    eventcounts[eventcount_128bitMMXuop] = (double) CounterValues[10];
    /* timestamp for the end of rotation 3 */
    timestamp[3] =(double) CounterValues[18];

    /* ROTATION 4*/
    /*************/
    define_events_rotate4();
    /* reset rotation 4 only events*/
    reset_event_counter(counters[8]);
    reset_event_counter(counters[10]);
    usleep(dt);
    get_event_counts(CounterValues);
    /* counts for rotation 4 only events*/
    eventcounts[eventcount_x87FPuop]        = (double) CounterValues[8];
    eventcounts[eventcount_x87SIMDmovesuop] = (double) CounterValues[10];
    /* counts for rotation 3&4 only events*/
    eventcounts[eventcount_FSBdataactivity] = (double) CounterValues[0];
    eventcounts[eventcount_MOBloadreplay]   = (double) CounterValues[2];
    eventcounts[eventcount_machineclear]     = (double) CounterValues[12];
    /* timestamp for the end of rotation 4 */
    timestamp[4] =(double) CounterValues[18];

    /* COUNTS FOR UNROTATED EVENTS: */
    eventcounts[eventcount_BSQcacheref]      = (double) CounterValues[1];
    eventcounts[eventcount_ITLBref]          = (double) CounterValues[3];
    eventcounts[eventcount_UOPqueuewrites07] = (double) CounterValues[4];
    eventcounts[eventcount_TCdelivermode]    = (double) CounterValues[5];
    eventcounts[eventcount_UOPqueuewrites04] = (double) CounterValues[6];
    eventcounts[eventcount_LDportreplay]     = (double) CounterValues[9];
    eventcounts[eventcount_STportreplay]     = (double) CounterValues[11];
    eventcounts[eventcount_UOPSretired]      = (double) CounterValues[13];
    eventcounts[eventcount_Frontend]         = (double) CounterValues[14];
    eventcounts[eventcount_UOPtype]          = (double) CounterValues[15];
    /* ALL COUNTER READIN ALG DONE NOW */
    /*------------------------------------------------------------------*/

    /* SEND EVENT COUNTS and TIMESTAMPS OVER ETHER */
    pos = buffer;
    for (i=0; i<NUM_EVENTS; i++)
    {
      pos += sprintf(pos, "%20.0lf ", eventcounts[i]);
      fprintf(stdout, "\n\n%d\n\n", (int) pos);
      /*21 char-s per element (20 FOR unsigned long long, 1 for space)*/
    }
    for (i=0; i<5; i++)
    {
      pos += sprintf(pos, "%20.0lf ", timestamp[i]);
            fprintf(stdout, "\n\n%d\n\n", (int) pos);
      /*21 char-s per element (20 FOR unsigned long long, 1 for space)*/
    }
    pos += sprintf(pos, "\n");
    socket_write(sock, buffer, strlen(buffer));
    /*-------------------------------------------------------------------*/

    /* POWER COMPUTATIONS */
    /**********************/
    /* DONE in Client now */

  }/*END OF WHILE(1)*/

}

int main(int argc, char **argv)
/* DOES SOCKET SETUPS, THEN CALLS PERFORMANCE READER */
{
  char hostname[MAX_CHAR];
  /* descriptions in Notes (24.1) */
  struct linger linger;
  struct hostent *resolve;
  struct sockaddr_in sock_addr = { 0 };
  int sock, status, on = 1;
  int portnum;

  /* get 1 cmd line params*/
  if (argc != 2)
  {
    fprintf(stderr, "Usage: %s <portnum>\n", argv[0]);
    exit(1);
  }

  portnum = atoi(argv[1]);/*convert to int*/

  sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sock == ERROR)
  {
    perror("socket()");
    exit(2);
  }

  status = setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (const char *) &on, sizeof(on));
  if (status == ERROR)
  {
    perror("setsockopt(...,SO_REUSEADDR,...)");
  }

  linger.l_onoff = 1; /* block socket close */
  linger.l_linger = 30; /* timeout [s]*/
  status = setsockopt(sock, SOL_SOCKET, SO_LINGER, (const char *) &linger, sizeof(linger));

  if (status == ERROR)
  {
    perror("setsockopt(...,SO_LINGER,...)");
  }

  gethostname(hostname, sizeof(hostname)); /* return the host's name this prog runnin */

  resolve = gethostbyname(hostname); /* get shortname for the host */
  if (resolve == NULL)
  {
    perror("gethostbyname()");
    exit(1);
  }

  memset(&sock_addr, 0, sizeof(sock_addr));
  memcpy(&sock_addr.sin_addr, resolve->h_addr, resolve->h_length);
  sock_addr.sin_family = AF_INET;
  sock_addr.sin_port = htons(portnum);

  fprintf(stdout, "\nServer %s: binding as %s\n", hostname, inet_ntoa(sock_addr.sin_addr));
  status = bind(sock, &sock_addr, sizeof(struct sockaddr_in));
  if (status == ERROR)
  {
    perror("bind()");
    exit(3);
  }

  fprintf(stdout, "Server: listening on port %d\n", portnum);
  status = listen(sock, BACKLOG);
  if (status == ERROR)
  {
    perror("listen()");
    exit(4);
  }
  /* now we have server listening for req. on specified port*/

  while(1)
  {
    struct sockaddr_in client_addr;
    int client_sock, client_len = sizeof(struct sockaddr_in);

    memset(&client_addr, 0, sizeof(struct sockaddr_in));
    client_sock = accept(sock, &client_addr, &client_len);
    if (client_sock == ERROR) {
      perror("accept()");
      exit(5);
    }

    status = getpeername(client_sock, &client_addr, &client_len);
    if (status == ERROR) {
      perror("getpeername()");
      exit(6);
    }

    fprintf(stdout, "Server: received request from %s\n",
    inet_ntoa(client_addr.sin_addr));

    PerformanceReader(client_sock);

    close(client_sock);
  }

  close(sock);

  return 0;
}











