/* Additional File For Performance Reader User Level code
 * Event definitions and ESCR & CCCR specifications are here
 * 
 * lowpowered by: canturk isci / canturk@princeton.edu / CANOO
 */

 /* THIS IS NOT A GENERIC ONE, DON'T USE AS TEMPLATE TO OTHERS*/

/* MSR ADDRESSES FOR THE 45 ESCR MSRs */
#define MSR_BSU_ESCR0   0x03A0
#define MSR_FSB_ESCR0   0x03A2
#define MSR_MOB_ESCR0   0x03AA
#define MSR_PMH_ESCR0   0x03AC
#define MSR_BPU_ESCR0   0x03B2
#define MSR_IS_ESCR0    0x03B4
#define MSR_ITLB_ESCR0  0x03B6
#define MSR_IX_ESCR0    0x03C8

#define MSR_BSU_ESCR1   0x03A1
#define MSR_FSB_ESCR1   0x03A3
#define MSR_MOB_ESCR1   0x03AB
#define MSR_PMH_ESCR1   0x03AD
#define MSR_BPU_ESCR1   0x03B3
#define MSR_IS_ESCR1    0x03B5
#define MSR_ITLB_ESCR1  0x03B7
#define MSR_IX_ESCR1    0x03C9

#define MSR_MS_ESCR0    0x03C0
#define MSR_TBPU_ESCR0  0x03C2
#define MSR_TC_ESCR0    0x03C4

#define MSR_MS_ESCR1    0x03C1
#define MSR_TBPU_ESCR1  0x03C3
#define MSR_TC_ESCR1    0x03C5

#define MSR_FIRM_ESCR0  0x03A4
#define MSR_FLAME_ESCR0 0x03A6
#define MSR_DAC_ESCR0   0x03A8
#define MSR_SAAT_ESCR0  0x03AE
#define MSR_U2L_ESCR0   0x03B0

#define MSR_FIRM_ESCR1  0x03A5
#define MSR_FLAME_ESCR1 0x03A7
#define MSR_DAC_ESCR1   0x03A9
#define MSR_SAAT_ESCR1  0x03AF
#define MSR_U2L_ESCR1   0x03B1

#define MSR_CRU_ESCR0   0x03B8
#define MSR_CRU_ESCR2   0x03CC
#define MSR_CRU_ESCR4   0x03E0
#define MSR_IQ_ESCR0    0x03BA
#define MSR_RAT_ESCR0   0x03BC
#define MSR_SSU_ESCR0   0x03BE
#define MSR_ALF_ESCR0   0x03CA

#define MSR_CRU_ESCR1   0x03B9
#define MSR_CRU_ESCR3   0x03CD
#define MSR_CRU_ESCR5   0x03E1
#define MSR_IQ_ESCR1    0x03BB
#define MSR_RAT_ESCR1   0x03BD
#define MSR_ALF_ESCR1   0x03CB

/* MSRs used for replay tagging*/
#define IA32_PEBS_ENABLE       0x3F1
#define MSR_PEBS_MATRIX_VERT   0x3F2

/* GLOBAL MSR ADDRESSES FOR THE 18 COUNTERS */
unsigned int counters[18] = {0x0300, 0x0301, 0x0302, 0x0303,/*0,1,2,3*/
                             0x0304, 0x0305, 0x0306, 0x0307,/*4,5,6,7*/
                             0x0308, 0x0309, 0x030A, 0x030B,/*8,9,10,11*/
                             0x030C, 0x030D, 0x030E, 0x030F,/*12,13,14,15*/
                             0x0310, 0x0311}; /*16,17*/

/* GLOBAL MSR ADDRESSES FOR THE 18 CCCR MSRs */
unsigned int CCCRs[18] = {0x0360, 0x0361, 0x0362, 0x0363,/*0,1,2,3*/
                          0x0364, 0x0365, 0x0366, 0x0367,/*4,5,6,7*/
                          0x0368, 0x0369, 0x036A, 0x036B,/*8,9,10,11*/
                          0x036C, 0x036D, 0x036E, 0x036F,/*12,13,14,15*/
                          0x0370, 0x0371}; /*16,17*/

struct ESCR_struct { unsigned int ESCR;       unsigned int Event_Select;
                     unsigned int Event_Mask; unsigned int Tag_Value;
                     unsigned int Tag_En;     unsigned int OS;
                     unsigned int USR;
                   } ESCR_Fields;

struct CCCR_struct { unsigned int CCCR;        unsigned int OVF;
                     unsigned int Cascade;     unsigned int OVF_PMI;
                     unsigned int Force_OVF;   unsigned int Edge;
                     unsigned int Threshold;   unsigned int Complement;
                     unsigned int Compare;     unsigned int Reserved_17_16;
                     unsigned int ESCR_Select; unsigned int EN;
                   } CCCR_Fields;

void define_events() /* For global events, that do not rotate */
{
  /* COUNTER 0: */
  /* NOT HERE */

  /* COUNTER 1: */
  /* EVENT: <BSQ_cache_reference> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_BSU_ESCR0; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x0C;
    ESCR_Fields.Event_Mask = 0x0507; /* count L2 RD hits shared,exc.,mod. */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[1];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x07;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 2: */
  /* NOT HERE */

  /* COUNTER 3: */
  /* EVENT: <ITLBE REFERENCE> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_ITLB_ESCR1; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x018;
    ESCR_Fields.Event_Mask = 0x07; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[3];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x03;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 4: */
  /* EVENT: <UOP_QUOEUE WRITES> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_MS_ESCR0; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x09;
    ESCR_Fields.Event_Mask = 0x07; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[4];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x00;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 5: */
  /* EVENT: <TC_DELIVER_MODE> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_TC_ESCR0; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x01;
    ESCR_Fields.Event_Mask = 0x038; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[5];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x01;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 6: */
  /* EVENT: <UOP_QUEUE_WRITES> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_MS_ESCR1; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x09;
    ESCR_Fields.Event_Mask = 0x04; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[6];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x0;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 7: */
  /* NEVER USED */

  /* COUNTER 8: */
  /* NOT HERE */

  /* COUNTER 9: */
  /* EVENT: <LD_PORT_REPLAY> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_SAAT_ESCR0; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x04;
    ESCR_Fields.Event_Mask = 0x02; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[9];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x02;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 10: */
  /* NOT HERE*/

  /* COUNTER 11: */
  /* EVENT: <STORE_PORT_REPLAY> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_SAAT_ESCR1; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x05;
    ESCR_Fields.Event_Mask = 0x02; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[11];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x02;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 12: */
  /* NOT HERE */

  /* COUNTER 13: */
    /* EVENT: <uops_retired> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_CRU_ESCR0; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x01;
    ESCR_Fields.Event_Mask = 0x03; /* count SPECULATIVE */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[13];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x04;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 14: */
  /* EVENT: < FRONT_END_EVENT> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_CRU_ESCR3; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x08;
    ESCR_Fields.Event_Mask = 0x03; /* count tagged & untagged & speculative */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[14];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x05;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 15: */
  /* EVENT: <uop type for front end tag> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_RAT_ESCR1; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x02;
    ESCR_Fields.Event_Mask = 0x06; /* tag only loads */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[15];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x02;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 16: */
  /* NEVER USED*/

  /* COUNTER 17: */
  /* NEVER USED */

}


void define_events_rotate1()
{
  /* COUNTER 0: */
  /* EVENT: <IOQ_ALLOCATION */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_FSB_ESCR0; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x03;
    ESCR_Fields.Event_Mask = 0x0EFE1; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[0];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x06;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 1: */
  /* NOT HERE */

  /* COUNTER 2: */
  /* EVENT: <BPU_FETCH_REQUESTS> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_BPU_ESCR1; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x03;
    ESCR_Fields.Event_Mask = 0x01; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[2];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x0;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 3: */
  /* NOT HERE */

  /* COUNTER 4: */
  /* NOT HERE */

  /* COUNTER 5: */
  /* NOT HERE */

  /* COUNTER 6: */
  /* NOT HERE */

  /* COUNTER 7: */
  /* NEVER USED*/

  /* COUNTER 8: */
  /* EVENT: <packed_sp_uop> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_FIRM_ESCR0; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x08;
    ESCR_Fields.Event_Mask = 0x08000; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[8];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x01;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 9: */
  /* NOT HERE */

  /* COUNTER 10: */
  /* EVENT: <PACKED_DP_UOP> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_FIRM_ESCR1; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x0C;
    ESCR_Fields.Event_Mask = 0x08000; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[10];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x01;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 11: */
  /* NOT HERE */

  /* COUNTER 12: */
  /* EVENT: <BRANCH_RETIRED> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_CRU_ESCR2; /* ESCR : MSR_CRU_ESCR2 */
    ESCR_Fields.Event_Select = 0x06;
    ESCR_Fields.Event_Mask = 0x0F; /* count only mispredicted */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[12];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x05;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers */
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 13: */
  /* NOT HERE */

  /* COUNTER 14: */
  /* NOT HERE */

  /* COUNTER 15: */
  /* NOT HERE */

  /* COUNTER 16: */
  /* NEVER USED */

  /* COUNTER 17: */
  /* NEVER USED */
}

void define_events_rotate2()
{
  /* COUNTER 0: */
  /* NOT HERE */

  /* COUNTER 1: */
  /* NOT HERE */

  /* COUNTER 2: */
  /* NOT HERE */

  /* COUNTER 3: */
  /* NOT HERE */

  /* COUNTER 4: */
  /* NOT HERE */

  /* COUNTER 5: */
  /* NOT HERE */

  /* COUNTER 6: */
  /* NOT HERE */

  /* COUNTER 7: */
  /* NEVER USED*/

  /* COUNTER 8: */
  /* EVENT: <SCALAR_SP_UOP> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_FIRM_ESCR0; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x0A;
    ESCR_Fields.Event_Mask = 0x08000; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[8];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x01;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 9: */
  /* NOT HERE */

  /* COUNTER 10: */
  /* EVENT: <SCALAR_DP_UOOP> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_FIRM_ESCR1; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x0E;
    ESCR_Fields.Event_Mask = 0x08000; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[10];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x01;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 11: */
  /* NOT HERE */

  /* COUNTER 12: */
  /* NOT HERE */

  /* COUNTER 13: */
  /* NOT HERE */

  /* COUNTER 14: */
  /* NOT HERE */

  /* COUNTER 15: */
  /* NOT HERE */

  /* COUNTER 16: */
  /* NEVER USED */

  /* COUNTER 17: */
  /* NEVER USED */
}

void define_events_rotate3()
{
  /* COUNTER 0: */
  /* EVENT: <FSB_DATA_ACTIVITY> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_FSB_ESCR0; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x017;
    ESCR_Fields.Event_Mask = 0x03F; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[0];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x06;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 1: */
  /* NOT HERE */

  /* COUNTER 2: */
  /* EVENT: <MOB_LOAD_REPLAY> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_MOB_ESCR1; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x03;
    ESCR_Fields.Event_Mask = 0x03A; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[2];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x02;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 3: */
  /* NOT HERE */

  /* COUNTER 4: */
  /* NOT HERE */

  /* COUNTER 5: */
  /* NOT HERE */

  /* COUNTER 6: */
  /* NOT HERE */

  /* COUNTER 7: */
  /* NEVER USED*/

  /* COUNTER 8: */
  /* EVENT: <64BIT_MMX_UOP> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_FIRM_ESCR0; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x02;
    ESCR_Fields.Event_Mask = 0x08000; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[8];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x01;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 9: */
  /* NOT HERE */

  /* COUNTER 10: */
  /* EVENT: <128BIT_MMX_UOP> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_FIRM_ESCR1; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x01A;
    ESCR_Fields.Event_Mask = 0x08000; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[10];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x01;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 11: */
  /* NOT HERE */

  /* COUNTER 12: */
  /* EVENT: <MACHINE_CLEAR> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_CRU_ESCR2; /* ESCR : MSR_CRU_ESCR2 */
    ESCR_Fields.Event_Select = 0x02;
    ESCR_Fields.Event_Mask = 0x01; /* count only mispredicted */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[12];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x05;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers */
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 13: */
  /* NOT HERE */

  /* COUNTER 14: */
  /* NOT HERE */

  /* COUNTER 15: */
  /* NOT HERE */

  /* COUNTER 16: */
  /* NEVER USED */

  /* COUNTER 17: */
  /* NEVER USED */
}

void define_events_rotate4()
{
  /* COUNTER 0: */
  /* MOT HERE */

  /* COUNTER 1: */
  /* NOT HERE */

  /* COUNTER 2: */
  /* NOT HERE */

  /* COUNTER 3: */
  /* NOT HERE */

  /* COUNTER 4: */
  /* NOT HERE */

  /* COUNTER 5: */
  /* NOT HERE */

  /* COUNTER 6: */
  /* NOT HERE */

  /* COUNTER 7: */
  /* NEVER USED*/

  /* COUNTER 8: */
  /* EVENT: <X87_FP_uop> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_FIRM_ESCR0; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x04;
    ESCR_Fields.Event_Mask = 0x08000; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[8];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x01;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 9: */
  /* NOT HERE */

  /* COUNTER 10: */
  /* EVENT: <X87_SIMD_MOVES_UOP> */
    /*ESCR Fields: */
    ESCR_Fields.ESCR = MSR_FIRM_ESCR1; /* ESCR : (IF 0, SYSCALL WILL IGNORE) */
    ESCR_Fields.Event_Select = 0x02E;
    ESCR_Fields.Event_Mask = 0x018; /* count all types */
    ESCR_Fields.Tag_Value = 0x0; /* No Tag*/
    ESCR_Fields.Tag_En = 0x0;
    ESCR_Fields.OS = 0x01; /* Count OS mode */
    ESCR_Fields.USR = 0x01; /* Count USR mode */
    /*CCCR Fields: */
    CCCR_Fields.CCCR = CCCRs[10];
    CCCR_Fields.OVF = 0x0;
    CCCR_Fields.Cascade = 0x0;
    CCCR_Fields.OVF_PMI = 0x0;
    CCCR_Fields.Force_OVF = 0x0;
    CCCR_Fields.Edge = 0x0;
    CCCR_Fields.Threshold = 0x00;
    CCCR_Fields.Complement = 0x0;
    CCCR_Fields.Compare = 0x0;
    CCCR_Fields.Reserved_17_16 = 0x03;
    CCCR_Fields.ESCR_Select = 0x01;
    CCCR_Fields.EN = 0X01; /* Don't Enable Yet*/
    /* Set the Registers
     * (DON'T CALL IF YOU DON'T YOU HAVEN'T DEFINED THE EVENT FOR THIS COUNTER)*/
    select_events(&ESCR_Fields,&CCCR_Fields);

  /* COUNTER 11: */
  /* NOT HERE */

  /* COUNTER 12: */
  /* NOT HERE */

  /* COUNTER 13: */
  /* NOT HERE */

  /* COUNTER 14: */
  /* NOT HERE */

  /* COUNTER 15: */
  /* NOT HERE */

  /* COUNTER 16: */
  /* NEVER USED */

  /* COUNTER 17: */
  /* NEVER USED */
}
